! (C) Copyright 2013 ECMWF.
!
! This software is licensed under the terms of the Apache Licence Version 2.0
! which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
! In applying this licence, ECMWF does not waive the privileges and immunities
! granted to it by virtue of its status as an intergovernmental organisation nor
! does it submit to any jurisdiction.

subroutine run() bind(c)
  use fckit_module
  character(len=:), allocatable :: name

  call fckit_main%name(name)
  call fckit_log%info("info: "//name)
  call fckit_log%debug("debug")
  if( .not. fckit_main%debug() ) then
    call fckit_log%warning("main debug not active")
  endif
end subroutine
